﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Alarm-Steuereinheit                                                                 //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2024 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Timer.h"
#include "Storing.h"
#include "AlarmControl.h"
#include "Cy4Net.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum							// Zustände des Alarmsystems:
{   STATE_STARTED              = 0,		// · Alarmsystem gestartet
	STATE_DISABLED             = 1,		// · Alarmsystem abgeschaltet
	STATE_ACTIVATION_ANNOUNCED = 2,		// · Alarmsystemaktivierung angekündigt
	STATE_ACTIVATED            = 3,		// · Alarmsystem aktiviert
	STATE_TRIGGER_ANNOUNCED    = 4,		// · Alarmauslösung angekündigt
	STATE_TRIGGERED            = 5		// · Alarm ausgelöst	
} State;

typedef enum							// Ereignisse:
{   EVENT_DEACTIVATE = 1,				// · Alarmsystem deaktivieren
	EVENT_ACTIVATE   = 2,				// · Alarmsystem aktivieren
	EVENT_TRIGGER    = 3,				// · Alarm auslösen
	EVENT_TIMEOUT    = 4				// · Zeitüberschreitung
} Event;

typedef struct                          // Ereignis-Satz:
{   UInt08 set;                         // · Satz mit Ereignissen vom Typ 'Event'
} EventSet;

typedef enum                            // Aktions-Codes des Alarmsystems für:
{   ACTION_CODE_REQUEST_STATE = 99,     // · Rundruf-Aufforderung an die Alarm-Steuereinheit
    ACTION_CODE_DEACTIVATION  =  1,     // · Deaktivierung des Alarmsystems
    ACTION_CODE_ACTIVATION	  =  3,     // · Aktivierung des Alarmsystems
    ACTION_CODE_TRIGGER_ALARM =  5      // · Alarmauslösung
} ActionCode;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static State     state             = STATE_STARTED;
static Timer	 stateMachineTimer = TIMER_SECONDS(0);
static Timer     stateRefreshTimer = TIMER_SECONDS(0);
static EventSet  eventSet          = { 0 };
static UInt08    durationA         = 0;
static UInt08    durationB         = 0;
static TextFeed  triggerEventText  = TEXTFEED("");

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt08
TriggerDuration(void)
{
	UInt08 duration;
	STORING_Load(STO_AlarmControlTriggerDuration, &duration);
	return duration;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichten-Ausstatter  //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupStateBroadcastMessage(Cy4Message* message)
{
	// Rundruf mit dem Zustand des Alarmsystems, siehe ALARMUNIT_StateBroadcast
    Cy4Address address = CY4ADDRESS_Create(CY4NET_Address().net, 0);
	UInt08 data[4] = { 0x20, 0xFF };
	data[2] = (UInt08)state;	
	switch (state)
	{   case STATE_ACTIVATION_ANNOUNCED: data[3] = durationA;		  break;
		case STATE_TRIGGER_ANNOUNCED:	 data[3] = durationB;		  break;
		case STATE_TRIGGERED:			 data[3] = TriggerDuration(); break;
		default:                         data[3] = 0xFF;
	}	
    CY4MESSAGE_SetupWrite(message, address, 0xE02000, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupDisabledPushMessage(Cy4Message* message)
{
	// Pushmessage, die bei Deaktivierung des Alarmsystem abgesetzt wird
    PushMessage pushMessage;
    STORING_Load(STO_AlarmControlEventPushMessage[0], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupActivationAnnouncedPushMessage(Cy4Message* message)
{
	// Pushmessage, die bei Ankündigung der Aktivierung des Alarmsystem abgesetzt wird
	PushMessage pushMessage;
	STORING_Load(STO_AlarmControlEventPushMessage[1], &pushMessage);
	PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupActivatedPushMessage(Cy4Message* message)
{
	// Pushmessage, die bei Aktivierung des Alarmsystem abgesetzt wird
    PushMessage pushMessage;
    STORING_Load(STO_AlarmControlEventPushMessage[2], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTriggerAnnouncedPushMessage(Cy4Message* message)
{
	// Pushmessage, die bei Ankündigung der Alarmauslösung abgesetzt wird
    PushMessage pushMessage;
    STORING_Load(STO_AlarmControlEventPushMessage[3], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTriggeredPushMessage(Cy4Message* message)
{
	// Pushmessage, die bei Alarmauslösung abgesetzt wird
    PushMessage pushMessage;
    STORING_Load(STO_AlarmControlEventPushMessage[4], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}
 
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupDisabledTextFeedMessage(Cy4Message* message)
{
	// Rundmitteilung, die bei Deaktivierung des Alarmsystem abgesetzt wird
    TextFeed textFeed;
    STORING_Load(STO_AlarmControlEventTextFeed[0], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupActivationAnnouncedTextFeedMessage(Cy4Message* message)
{
	// Rundmitteilung, die bei Ankündigung der Aktivierung des Alarmsystem abgesetzt wird
    TextFeed textFeed;
    STORING_Load(STO_AlarmControlEventTextFeed[1], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupActivatedTextFeedMessage(Cy4Message* message)
{
	// Rundmitteilung, die bei Aktivierung des Alarmsystem abgesetzt wird
    TextFeed textFeed;
    STORING_Load(STO_AlarmControlEventTextFeed[2], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTriggerAnnouncedTextFeedMessage(Cy4Message* message)
{
	// Rundmitteilung, die bei Ankündigung der Alarmauslösung abgesetzt wird
	TextFeed textFeed;
	STORING_Load(STO_AlarmControlEventTextFeed[3], &textFeed);
	TEXTFEED_ToMessage(&textFeed, message);
}
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTriggeredTextFeedMessage(Cy4Message* message)
{	
	// Rundmitteilung, die bei Alarmauslösung abgesetzt wird
	TEXTFEED_ToMessage(&triggerEventText, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SendStateBroadcast(void)
{
    if (!CY4NET_IsPending(SetupStateBroadcastMessage, NULL))
    {   CY4NET_Post(SetupStateBroadcastMessage, NULL);		
    }
	TIMER_Restart(&stateRefreshTimer);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
AddEvent(UInt08 eventNo)
{
    if( eventNo > 0 ) eventSet.set |= BV(eventNo-1);
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
HasEvent(UInt08 eventNo)
{
    return eventNo > 0 && eventSet.set & BV(eventNo-1);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ClearEvents(void)
{
    eventSet.set = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfrage-Bearbeiter  ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_AlarmSystemInitialState (read write)                           //
//                                                                                      //
//  Beschreibung: Bestimmt den Aktivierungszustand des Alarmsystem nach Systemstart.	//
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |IST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [IST] Aktivierungszustand nach Systemstart: 0 = Deaktiviert, 1 = Aktiviert, 2 =		//
//        Wie zuletzt, x = keine Funktion.												//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_AlarmSystemInitialState(const Cy4Service* service, Cy4Message* message)
{
	AlarmSystemInitialState initialState;
	STORING_Load(STO_AlarmSystemInitialState, &initialState);
	UInt08 data[1] = { (UInt08)initialState };
	CY4SERVICE_TransferPayload(service, message, data);	
	if (data[0] < 3)
	{	initialState = (AlarmSystemInitialState)data[0];
		STORING_Save(&initialState, STO_AlarmSystemInitialState);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_AlarmTriggerDuration (read write)                              //
//                                                                                      //
//  Beschreibung: Zeit, nach der ein ausgelöster Alarm wieder abgeschaltet wird.        //
//                Nach Ablauf von 'DUR' Minuten wechselt das Alarmsystem vom Zustand	//
//                'STATE_TRIGGERED' zurück in den Zustand 'STATE_ACTIVATED', d.h. ein   //
//                weiterer Auslöser würde einen erneuten Alarm auslösen.				//
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |DUR|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [DUR] Dauer der Alarmauslösung: 1 .. 60 min.                                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_AlarmTriggerDuration(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[1];
    data[0] = TriggerDuration();
    CY4SERVICE_TransferPayload(service, message, data);
    data[0] = Bound(1, data[0], 60);
    STORING_Save(data, STO_AlarmControlTriggerDuration);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_EventPushMessage (read write)                                  //
//                                                                                      //
//  Beschreibung: Push-Nachricht, die bei Eintritt des jeweiligen Ereignisses versendet //
//                wird:                                                                 //
//                                                                                      //
//       A01000h:6: Push-Nachricht bei Ereignis: Alarmsystem abgeschaltet				//
//       A01100h:6: Push-Nachricht bei Ereignis: Alarmsystemaktivierung angekündigt     //
//       A01200h:6: Push-Nachricht bei Ereignis: Alarmsystem aktiviert					//
//       A01300h:6: Push-Nachricht bei Ereignis: Alarmauslösung angekündigt				//
//       A01400h:6: Push-Nachricht bei Ereignis: Alarm ausgelöst						//
//                                                                                      //
//               0   1   2   3   4   5                                                  //
//   Belegung: +---+---+---+---+---+---+                                                //
//             |NET|DEV|LC0|LC1|LC2|PAR|                                                //
//             +---+---+---+---+---+---+                                                //
//                                                                                      //
//  [NET] Push-Nachricht Ziel bei o.g. Ereignis, Netzadresse   : 0 ..     255           //
//  [DEV] Push-Nachricht Ziel bei o.g. Ereignis, Geräteadresse : 0 ..     255           //
//  [LCx] Push-Nachricht Ziel bei o.g. Ereignis, Dienstort     : 0 .. 0xFFFFF           //
//  [PAR] Push-Nachricht Ziel bei o.g. Ereignis, Parameter-Byte: 0 ..     255           //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_EventPushMessage(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 index;
    if     ( location < 0xA01100 ) index = 0;
    else if( location < 0xA01200 ) index = 1;
    else if( location < 0xA01300 ) index = 2;
	else if( location < 0xA01400 ) index = 3;
    else                           index = 4;
    PushMessage pushMessage;
    STORING_Load(STO_AlarmControlEventPushMessage[index], &pushMessage);
    UInt08  data[6];
    PUSHMESSAGE_Serialize(&pushMessage, data);
    CY4SERVICE_TransferPayload(service, message, data);
    PUSHMESSAGE_Deserialize(data, &pushMessage);
    STORING_Save(&pushMessage, STO_AlarmControlEventPushMessage[index]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_EventTextFeed (read write)                                     //
//                                                                                      //
//  Beschreibung: Rundmitteilung, die bei Eintritt des jeweiligen Ereignisses versendet //
//                wird:                                                                 //
//                                                                                      //
//        A02000h:51: Rundmitteilung bei Ereignis: Alarmsystem abgeschaltet				//
//        A02100h:51: Rundmitteilung bei Ereignis: Alarmsystemaktivierung angekündigt   //
//        A02200h:51: Rundmitteilung bei Ereignis: Alarmsystem aktiviert				//
//        A02300h:51: Rundmitteilung bei Ereignis: Alarmauslösung angekündigt           //
//                                                                                      //
//                Bei Alarmauslösung wird der von der auslösenden Alarmeinheit mitge-   //
//                lieferte Rundmitteilungstext ausgegeben (s. ALARMCONTROL_Trigger-		//
//                Alarm):                                                               //
//                                                                                      //
//               0   1   2        49  50                                                //
//             +---+---+---+-   -+---+---+                                              //
//   Belegung: |T00|T01|T02| ... |T49| 0 |                                              //
//             +---+---+---+-   -+---+---+                                              //
//                                                                                      //
//                                                                                      //
//  [Txx] Zeichen der Rundmitteilung bei o.g. Ereignis: 0 .. 255                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_EventTextFeed(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 index;
    if     ( location < 0xA02100 ) index = 0;
    else if( location < 0xA02200 ) index = 1;
	else if( location < 0xA02300 ) index = 2;
    else                           index = 3;
    TextFeed textFeed;
    STORING_Load(STO_AlarmControlEventTextFeed[index], &textFeed);
    UInt08  data[51];
    TEXTFEED_Serialize(&textFeed, data);
    CY4SERVICE_TransferPayload(service, message, data);
    TEXTFEED_Deserialize(data, &textFeed);
    STORING_Save(&textFeed, STO_AlarmControlEventTextFeed[index]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_RequestStateBroadcast (write only)                             //
//                                                                                      //
//  Beschreibung: Initiiert die Alarm-Steuereinheit dazu, allen Teilnehmern des Alarm-  //
//                verbundes mittels Rundruf den offiziellen Zustand des Alarmsystems 	//
//                mitzuteilen.															//
//				  Die Schreibaufforderung ist nur gültig, wenn [TRx] den unten ange-    //
//                gebene Auslösewert und [COD] und den Aktionscode mitführt.			//
//                                                                                      //
//              0   1   2                                                               //
//  Belegung: +---+---+---+                                                             //
//            |TR1|TR2|COD|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [TRx] Auslösewert: 20FFh (8447).												    //
//  [COD] Code für Rundruf-Aufforderung: 99.											//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_RequestStateBroadcast(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[3] = { 0, 0, 0 };	
    CY4SERVICE_TransferPayload(service, message, data);
	if (ToWord(data[0], data[1]) == 0x20FF && data[2] == (UInt08)ACTION_CODE_REQUEST_STATE)
	{	SendStateBroadcast();
	}
	else                                     
	{	CY4MESSAGE_Invalidate(message);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_Activate (write only)                                          //
//                                                                                      //
//  Beschreibung: Aktiviert das Alarmsystem.											//
//				  Die Schreibaufforderung ist nur gültig, wenn [TRx] den unten ange-    //
//                gebene Auslösewert und [COD] den u.a. Aktionscode mitführt.			//
//                Bevor das Alarmsystem in den Zustand 'STATE_ACTIVATED' wechselt, ver-	//
//                weilt das System noch [DUR] Sekunden im Zustand 'STATE_ACTIVATION_-	//
//				  ANNOUNCED'. Die zeitversetzte Aktivierung erlaubt beispielsweise nach //
//				  Aktivierung des Alarmsystems das Haus zu verlassen ohne dass das		//
//				  Öffnen der Haustür unmittelbar einen Alarm auslöst.					//
//                Wird [DUR] nicht beschrieben erhält [DUR] den Wert 0.					//
//                                                                                      //
//              0   1   2   3                                                           //
//            +---+---+---+---+                                                         //
//  Belegung: |TR1|TR2|COD|DUR|                                                         //
//            +---+---+---+---+                                                         //
//                                                                                      //
//  [TRx] Auslösewert: 20FFh (8447).												    //
//  [COD] Code für die Aktivierung: 3.													//
//  [DUR] Dauer der Aktivierungsankündigung: 0 .. 240 [s]. (0, wenn ungenutzt)          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_Activate(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[4] = { 0, 0, 0, 0 };    	
    CY4SERVICE_TransferPayload(service, message, data);
	if (ToWord(data[0], data[1]) == 0x20FF && data[2] == (UInt08)ACTION_CODE_ACTIVATION)
	{	AddEvent(EVENT_ACTIVATE);
		durationA = Bound(0, data[3], 240);
	}
	else                                     
	{	CY4MESSAGE_Invalidate(message);
	}	
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_Deactivate (write only)                                        //
//                                                                                      //
//  Beschreibung: Deaktiviert das Alarmsystem ab.                                       //
//				  Die Schreibaufforderung ist nur gültig, wenn [TRx] den unten ange-    //
//                gebene Auslösewert und [COD] den u.a. Aktionscode mitführt.		    //
//                                                                                      //
//              0   1   2                                                               //
//            +---+---+---+                                                             //
//  Belegung: |TR1|TR2|COD|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [TRx] Auslösewert: 20FFh (8447).												    //
//  [COD] Code für die Deaktivierung: 1.												//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_Deactivate(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[3] = { 0, 0, 0 };	
	CY4SERVICE_TransferPayload(service, message, data);
	if (ToWord(data[0], data[1]) == 0x20FF && data[2] == (UInt08)ACTION_CODE_DEACTIVATION)
	{	AddEvent(EVENT_DEACTIVATE);        
	}
	else                                     
	{	CY4MESSAGE_Invalidate(message);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMCONTROL_TriggerAlarm (write only)                                      //
//                                                                                      //
//  Beschreibung: Rundruf für das Auslösen eines Alarms.								//
//				  Die Schreibaufforderung ist nur gültig, wenn [TRx] den unten ange-    //
//                gebene Auslösewert und [COD] den entspr. Aktionscode mitführt.		//
//				  Voraussetzung für das Auslösen eines Alarms ist, dass sich das Alarm-	//
//				  system im Zustand 'STATE_ACTIVATED' befindet. In anderen Zuständen	//
//				  bleibt das Auslösen eines Alarms ohne Wirkung.						//
//                Bevor das Alarmsystem in den Zustand 'STATE_TRIGGERED' wechselt, ver-	//
//                weilt das System noch [DUR] Sekunden im Zustand 'STATE_TRIGGER_-	    //
//				  ANNOUNCED'. Die zeitversetzte Auslösung erlaubt die Deaktivierung des //
//				  Alarmsystems obwohl bereits ein Alarm ausgelöst wurde, z.B. nach dem	//		
//				  öffnen der Haustür.													//
//				  Die auslösende Alarmeinheit kann in [Txx] einen Rundmitteilungstext	//
//				  mitliefern, der am Falle der Alarmauslösung, also bei Wechsel in den	//
//				  Zustand 'STATE_TRIGGERED', versendet wird.							//
//				  Werden [DUR] bzw. [Txx] nicht beschrieben erhält [DUR] den Wert 0,	//
//				  und [Txx] bleibt leer (T00 = '0'), d.h. es findet kein Textversendung //
//				  statt.																//
//                                                                                      //
//              0   1   2   3   4   5   6        53  54                                 //
//            +---+---+---+---+---+---+---+-   -+---+---+                               //
//  Belegung: |TR1|TR2|COD|DUR|T00|T01|T02| ... |T49| 0 |                               //
//            +---+---+---+---+---+---+---+-   -+---+---+                               //
//                                                                                      //
//  [TRx] Auslösewert: 20FFh (8447).												    //
//  [COD] Code für Alarmauslösung: 5.													//
//  [DUR] Dauer der Alarm-Auslöseankündigung: 0 .. 240 [s] (0).						    //
//  [Txx] Zeichen der Rundmitteilung bei Auslösung des Alarms: 0 .. 255 (T00 = 0).		//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMCONTROL_TriggerAlarm(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[55] = { 0, 0, 0, 0, 0 };
	CY4SERVICE_TransferPayload(service, message, data);			
	if (ToWord(data[0], data[1]) == 0x20FF && data[2] == (UInt08)ACTION_CODE_TRIGGER_ALARM)
	{	AddEvent(EVENT_TRIGGER);
		durationB = Bound(0, data[3], 240);
		TEXTFEED_Deserialize(&data[4], &triggerEventText);
	}
	else
	{	CY4MESSAGE_Invalidate(message);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//   Zustandsautomat Alarm-Steuereinheit:                                               //
//                                                                                      //
//                                      (O)                                             //
//                                       | StartTimer(CFG_ALARM_SYSTEM_STARTUP_DELAY)   //
//                                       |                                              //
//                                      \|/                                             //
//                           +-----------------------+  EventTimeout (und Initial Ein)	//
//                           |                       |  -> SendStateBroadcast()			//
//                           |        STARTED        +---------------------------+      //
//                           |                       |                           |      //
//                           +-----------+-----------+                           |      //
//                                       | EventTimeout (und Initial Aus)        |      //
//                                       | -> SendStateBroadcast()               |      //
//                                       |                                       |      //
//                                      \|/                                      |      //
//                           +------------------------+                          |      //
//         +---------------->|                        |                          |      //
//         |                 |        DISABLED        |                          |      //
//         |<----------------+                        |                          |      //
//         |   EventDisable  +-----------+------------+                          |      //
//         |   -> SendStateBroadcast()   | EventActivate                         |      //
//         |                             | -> StartTimer(a)                      |      //
//         |                             | -> SendStateBroadcast()               |      //
//         |                             |                                       |      //
//         |                             |<--------------------+                 |      //
//         |                             |                     |                 |      //
//         |                            \|/                    |                 |      //
//         |                 +------------------------+        |                 |      //
//         |                 |                        |        |                 |      //
//         |                 |  ACTIVATION_ANNOUNCED  +--------+                 |      //
//         |<----------------+                        | EventActivate            |      //
//         |   EventDisable  +-----------+------------+ -> StartTimer(a)         |      //
//         |   -> SendStateBroadcast()   | EventTimeout                          |      //
//         |                             | -> SendStateBroadcast()               |      //
//         |                             |                                       |      //
//         |                            \|/                                      |      //
//         |                 +------------------------+                          |      //
//         |                 |                        |<-------------------------+      //
//         |                 |       ACTIVATED        |                                 //
//         |<----------------+                        |<-------------------------+      //
//         |   EventDisable  +-----------+------------+                          |      //
//         |   -> SendStateBroadcast()   | EventTrigger                          |      //
//         |                             | -> StartTimer(b)                      |      //
//         |                             | -> SendStateBroadcast()               |      //
//         |                             |                                       |      //
//         |                            \|/                                      |      //
//         |                +-------------------------+                          |      //
//         |                |                         |                          |      //
//         |                |    TRIGGER_ANNOUNCED    +------------------------->|      //
//         |<---------------+                         | EventActivate            |		//
//         |  EventDisable  +------------+------------+ -> SendStateBroadcast()  |      //
//         |  -> SendStateBroadcast()    | EventTimeout                          |      //
//         |                             | -> SendStateBroadcast()               |      //
//         |                             |                                       |      //
//         |                            \|/                                      |      //
//         |                +-------------------------+                          |      //
//         |                |                         |                          |      //
//         |                |        TRIGGERED        +--------------------------+      //
//         +----------------+                         | EventTimeout, EventActivate 	//
//            EventDisable  +-------------------------+	-> SendStateBroadcast()		    //
//            -> SendStateBroadcast()                                                   //
//                                                                                      //
//      a = Dauer der Aktivierungsankündigung                                           //
//      b = Dauer der Auslösungsankündigung                                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedStateControl(void)
{	
    // Ereignisgenerator (alle außer EVENT_TIMEOUT werden außerhalb der Funkt. ausgelöst)
    if (TIMER_HasExpired(&stateMachineTimer)) AddEvent(EVENT_TIMEOUT);
    // Zustandsübergang	
    switch (state)
    {   case STATE_STARTED:
            if (HasEvent(EVENT_TIMEOUT))
            {   bool enabled;
				AlarmSystemInitialState initialState;
				STORING_Load(STO_AlarmSystemInitialState, &initialState);				
				switch (initialState)
				{	case ALARM_SYSTEM_INITIAL_DISABLED: enabled = false; break;
					case ALARM_SYSTEM_INITIAL_ENABLED:  enabled = true; break;
					default:						    STORING_Load(STO_AlarmSystemEnabled, &enabled);
				}                
                if (enabled) state = STATE_ACTIVATED;
                else         state = STATE_DISABLED;
				SendStateBroadcast();
            }
            break;
        case STATE_DISABLED:
            if (HasEvent(EVENT_ACTIVATE))
            {   if (durationA == 0)
				{	// Keine Vorankündigung, Alarmsystem sofort aktivieren
					state = STATE_ACTIVATED;
				}
				else
				{	TIMER_Start(&stateMachineTimer, (UInt16)durationA);
					state = STATE_ACTIVATION_ANNOUNCED;
				}
				SendStateBroadcast();
            }            
            else if (HasEvent(EVENT_DEACTIVATE))
            {   SendStateBroadcast();
            }
            break;
        case STATE_ACTIVATION_ANNOUNCED:
            if (HasEvent(EVENT_DEACTIVATE))
            {   state = STATE_DISABLED;
				SendStateBroadcast();
            }            
            else if (HasEvent(EVENT_ACTIVATE))
            {   TIMER_Start(&stateMachineTimer, (UInt16)durationA);                
            }
            else if (HasEvent(EVENT_TIMEOUT))
            {   state = STATE_ACTIVATED;
				SendStateBroadcast();
            }
            break;
        case STATE_ACTIVATED:
            if (HasEvent(EVENT_DEACTIVATE))
            {   state = STATE_DISABLED;
				SendStateBroadcast();
            }
            else if (HasEvent(EVENT_TRIGGER))            
            {   if (durationB == 0)
				{	// Keine Vorankündigung, Alarm sofort auslösen
					UInt08 duration = TriggerDuration();
					TIMER_Start(&stateMachineTimer, (UInt16)duration*60);
					state = STATE_TRIGGERED;
				}
				else
				{	TIMER_Start(&stateMachineTimer, (UInt16)durationB);                
					state = STATE_TRIGGER_ANNOUNCED;
				}
				SendStateBroadcast();
            }
            break;
        case STATE_TRIGGER_ANNOUNCED:
            if (HasEvent(EVENT_DEACTIVATE))
            {   state = STATE_DISABLED;
				SendStateBroadcast();
            }                                
			else if (HasEvent(EVENT_ACTIVATE))
            {   state = STATE_ACTIVATED;
				SendStateBroadcast();
            }
            else if (HasEvent(EVENT_TIMEOUT))
            {   UInt08 duration = TriggerDuration();
                TIMER_Start(&stateMachineTimer, (UInt16)duration*60);
                state = STATE_TRIGGERED;
				SendStateBroadcast();
            }
            break;
        case STATE_TRIGGERED:
            if(HasEvent(EVENT_TIMEOUT) || HasEvent(EVENT_ACTIVATE))
			{	state = STATE_ACTIVATED;
				SendStateBroadcast();
			}
            else if (HasEvent(EVENT_DEACTIVATE))
			{	state = STATE_DISABLED;
				SendStateBroadcast();
			}			
            break;
        default:        	
			BOARD_Reset();
    }
	ClearEvents();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedRefreshStateBroadcast(void)
{
	if (TIMER_HasExpired(&stateRefreshTimer))
	{	SendStateBroadcast();
	}	
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedStateChangedNotifications(void)
{
    UInt08 index = (UInt08)state-1;
    // Ereignis-Ausgaben
    static State prevState = STATE_STARTED;
    if (prevState != state)
    {   prevState  = state;
        // Push-Nachricht
        {   static const Cy4MessageSetup  messageSetupTable[] PROGMEM =
            {   SetupDisabledPushMessage,				   
				SetupActivationAnnouncedPushMessage,	   
                SetupActivatedPushMessage,				   
                SetupTriggerAnnouncedPushMessage,		   
                SetupTriggeredPushMessage				   
            };
            Cy4MessageSetup messageSetup = (Cy4MessageSetup)pgm_read_word(&messageSetupTable[index]);
            if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);
        }
        // Rundmitteilung        
        static const Cy4MessageSetup  messageSetupTable[] PROGMEM =
        {   SetupDisabledTextFeedMessage,			   
			SetupActivationAnnouncedTextFeedMessage,   
			SetupActivatedTextFeedMessage,			   
			SetupTriggerAnnouncedTextFeedMessage,	   
			SetupTriggeredTextFeedMessage			   
        };
        Cy4MessageSetup messageSetup = (Cy4MessageSetup)pgm_read_word(&messageSetupTable[index]);
        if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);        
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedActivationStateSaving(void)
{	
	AlarmSystemInitialState initialState;
	STORING_Load(STO_AlarmSystemInitialState, &initialState);
	if (initialState == ALARM_SYSTEM_INITIAL_AS_LAST)
	{	// Aktivierungszustand speichern		
		bool enabled;
		STORING_Load(STO_AlarmSystemEnabled, &enabled);
		if ((state == STATE_DISABLED && enabled) || (state == STATE_ACTIVATED && !enabled))
		{   enabled = state == STATE_ACTIVATED;
			STORING_Save(&enabled, STO_AlarmSystemEnabled);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMCONTROL_Startup(void)
{    	
	TIMER_Start(&stateMachineTimer, CFG_ALARM_CONTROL_STARTUP_DELAY);
	TIMER_Start(&stateRefreshTimer, CFG_ALARM_CONTROL_STATE_CONFIRMATION_INTERVAL*60);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMCONTROL_Proceed(void)
{

    ProceedStateControl();	
    ProceedRefreshStateBroadcast();
    ProceedStateChangedNotifications();
	ProceedActivationStateSaving();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMCONTROL_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
